/**
 *   wsdl-common-jbi-binding - Class(es) for supporting JBI WSDL deployments
 *
 *   Copyright (C) 2007 Gestalt, LLC. All Rights Reserved.
 *   http://www.gestalt-llc.com/
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License version 2.1 as published by the Free Software Foundation.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
 */

package com.gestalt.jbi.component.manager.deployment;

import com.gestalt.jbi.component.AbstractComponent;
import com.gestalt.jbi.component.descriptor.JBIDescriptorParser;
import com.gestalt.jbi.component.manager.ServiceUnit;
import org.w3c.dom.Document;

import javax.jbi.management.DeploymentException;
import javax.jbi.messaging.MessageExchange;
import static javax.jbi.messaging.MessageExchange.Role.CONSUMER;
import static javax.jbi.messaging.MessageExchange.Role.PROVIDER;
import javax.wsdl.Binding;
import javax.wsdl.Definition;
import javax.wsdl.Port;
import javax.wsdl.Service;
import javax.wsdl.WSDLException;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.ExtensionRegistry;
import javax.wsdl.factory.WSDLFactory;
import javax.wsdl.xml.WSDLReader;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

public abstract class AbstractWSDLDeployer extends AbstractDeployer {

    /**
     * The log for this class and subclasses.
     * <p>
     * <b>Note to developers:</b>
     * Whether or not it is good practice to use inherited loggers,
     * some subclasses of this class <i>do</i> expect to inherit this,
     * so leave it either public or protected.
     */
    protected static final Logger log =
          Logger.getLogger (AbstractWSDLDeployer.class.getName());

    protected static WSDLFactory wsdlFactory;
    protected static DocumentBuilderFactory documentFactory =
                DocumentBuilderFactory.newInstance();
    protected static DocumentBuilder documentBuilder;

    public AbstractWSDLDeployer(AbstractComponent component) {
        super(component);
        try {
            wsdlFactory = WSDLFactory.newInstance();
            documentBuilder = documentFactory.newDocumentBuilder();
        } catch (Exception e) {
            log.throwing(getClass().getName(), "init", e);
        }
    }

    public boolean canDeploy(String serviceUnitRootPath) {
        List<File> wsdls = scanDirectory(new File(serviceUnitRootPath));
        log.fine("We canDeploy was found to be - "
                + ((wsdls == null ? 0 : wsdls.size()) > 0));
        return (wsdls == null ? 0 : wsdls.size()) > 0;
    }

    private List<File> scanDirectory(File directory) {
        File[] files = directory.listFiles();
        List<File> tmp = new ArrayList<File>();
        for (File file : files) {
            if (file.isDirectory()) {
                tmp.addAll(scanDirectory(file));
            } else {
                if (file.getName().endsWith(".wsdl")) {
                    tmp.add(file);
                }
            }
        }
        return tmp;
    }

    public ServiceUnit deploy(String serviceUnitName,
                              String serviceUnitRootPath)
            throws DeploymentException {
        List<File> wsdls = scanDirectory(new File(serviceUnitRootPath));
        if ((wsdls == null ? 0 : wsdls.size()) == 0) {
            log.warning("Failed to deploy service unit - no WSDL files found - "
                    + serviceUnitName);
            throw new DeploymentException("Failed to deploy service unit - " +
                    "no WSDL files found - " + serviceUnitName);
        }
        ServiceUnit serviceUnit = createServiceUnit(serviceUnitName,
                serviceUnitRootPath);
        JBIDescriptorParser descriptorParser =
                new JBIDescriptorParser(serviceUnitRootPath);
        for (File wsdl : wsdls) {
            try {
                log.fine("Trying to parse a WSDL - " + wsdl.getAbsolutePath());
                parseWSDL(serviceUnit, wsdl, descriptorParser);
            } catch (Exception e) {
                log.log (Level.WARNING, "Cannot parse WSDL", e);
            }
        }
        if (serviceUnit.getEndpoints().size() == 0) {
            log.warning("Failed to deploy service unit - no endpoints found - "
                    + serviceUnitName);
            throw new DeploymentException("Failed to deploy service unit - " +
                    "no endpoints found - " + serviceUnitName);
        }
        return serviceUnit;
    }

    public Definition getWsdlDefinition(File wsdl) throws WSDLException {
        WSDLReader reader = wsdlFactory.newWSDLReader();
        reader.setExtensionRegistry(createExtensionRegistry());
        return reader.readWSDL(wsdl.getAbsolutePath());
    }

    public Document getWsdlDocument(File wsdl) throws Exception {
        return documentBuilder.parse(wsdl);
    }

    protected Binding getBinding(Definition definition, String serviceName,
                                 String endpointName) {
        Port port = getPort(definition, serviceName, endpointName);
        if (port == null) {
            log.warning("No port found for binding");
            return null;
        }
        return port.getBinding();
    }

    protected Port getPort(Definition definition, String serviceName,
                           String endpointName) {
        log.fine("Looking for serviceName - " + serviceName);
        Service svc = definition.getService(QName.valueOf(serviceName));
        if (svc == null) {
            log.warning ("Service " + serviceName + " not found");
            return null;
        }
        log.fine("Looking for endpointName - " + endpointName);
        Port port = svc.getPort(QName.valueOf(endpointName).getLocalPart());
        if (port == null) {
            log.warning("No port found");
            log.warning ("Port " + endpointName + " not found");
            return null;
        }
        return port;
    }

    /**
     * Returns the concrete binding
     */
    protected <T extends ExtensibilityElement> T getConcreteBinding(Definition definition,
            String serviceName, String endpointName, Class<T> bindingClass) {
        Binding binding = getBinding(definition, serviceName, endpointName);
        if (binding != null) {
            List<ExtensibilityElement> extElems =
                    binding.getExtensibilityElements();
            if (extElems != null) {
                for (ExtensibilityElement extensibilityElement : extElems) {
                    try {
                        if (extensibilityElement.getClass().
                              isAssignableFrom (bindingClass)) {
                            return (T) extensibilityElement;
                        }
                    } catch (Exception e) {
                        log.warning("Throwing " + e);
                        return null;
                    }
                }
            }
        }
        return null;
    }

    /**
     * Returns the concrete address
     * Want to use it like
     * RSSAddress address = getConcreteAddress(def, sn, en, RSSAddress.class)
     * UDDIAddress address = getConcreteAddress(def, sn, en, UDDIAddress.class)
     */
    protected <T extends ExtensibilityElement> T getConcreteAddress(Definition definition,
            String serviceName, String endpointName, Class<T> compareTo) {
        Port port = getPort(definition, serviceName, endpointName);
        if (port != null) {
            List<ExtensibilityElement> extensibilityElements =
                    port.getExtensibilityElements();
            if (extensibilityElements != null) {
                for (ExtensibilityElement extensibilityElement : extensibilityElements) {
                    if (extensibilityElement != null) {
                        try {
                            if (extensibilityElement.getClass().
                                    isInstance(compareTo.newInstance())) {
                                return (T) extensibilityElement;
                            }
                        } catch (Exception e) {
                            log.warning("Throwing " + e);
                            return null;
                        }
                    }
                }
            }
        }
        return null;
    }

    protected ServiceUnit createServiceUnit(String serviceUnitName,
            String serviceUnitRootPath) {
        return new ServiceUnit(serviceUnitName, serviceUnitRootPath, component);
    }

    /**
     * Handles the parsing of the WSDL and will create the endpoints for
     * providers and consumers,
     */
    protected void parseWSDL(ServiceUnit serviceUnit, File wsdl,
            JBIDescriptorParser descriptorParser) throws Exception {
        log.fine("Parsing the WSDL: " + wsdl);
        for (JBIDescriptorParser.Consumer consumer : descriptorParser.getConsumers()) {
            addEndpoint(serviceUnit, wsdl, consumer.getServiceName(), consumer.getEndpointName(),
                    consumer.getInterfaceName(), CONSUMER);
        }
        for (JBIDescriptorParser.Provider provider : descriptorParser.getProviders()) {

            addEndpoint(serviceUnit, wsdl, provider.getServiceName(), provider.getEndpointName(),
                    provider.getInterfaceName(), PROVIDER);
        }
        log.fine("Processing Complete");
    }

    protected abstract void addEndpoint(ServiceUnit serviceUnit, File wsdl,
            QName serviceName, String endpointName, QName interfaceName, MessageExchange.Role role)
            throws Exception;

    /**
     * Creates the Concrete ExtensionRegistry class used when
     * building the WSDL Factory to register wsdl extensions.
     * @return The concrete ExtensionRegistry such as a
     * <code>new RSSExtensionRegistry()</code>
     */
    public abstract ExtensionRegistry createExtensionRegistry();

}
